from langutils.langs import base_grammar
from .huruf import huruf
from .item_block import (class_item, exception_item, function_item,
                        interface_item, scope_item)
from .item_export import export_item
from .item_import import import_item
from .item_package import package_item
from .item_statement import statement_item
from .keywords import keywords

from .searchable import searchable
from .zops.dataops import dataops_statement

# from zops.fmus_ops import fmus_ops

# sekarang masih lang
# nanti dukung: library, framework, package manager, docker/kubernetes/cloud/terraform
# juga hrs dukung: name alias = symtab, type alias
bahasa = f"""
program: program_configuration? item_line+

//program_configuration: "/" program_configuration_items "/"
program_configuration: program_configuration_items "/"
program_configuration_items: program_configuration_item ("," program_configuration_item)*
program_configuration_item: language_context
  | library_context
  | framework_context
  | tools_context

language_context: "py" -> py
  | "clj"     -> clj
  | "cpp"     -> cpp
  | "cs"      -> cs
  | "dart"    -> dart
  | "elixir"  -> elixir
  | "erlang"  -> erlang
  | "go"      -> go
  | "hs"      -> hs
  | "java"    -> java  
  | "js"      -> js
  | "kt"      -> kt
  | "php"     -> php
  | "pl"      -> pl
  | "rb"      -> rb
  | "rs"      -> rs
  | "scala"   -> scala
  | "swift"   -> swift
  | "ts"      -> ts
  | "v"       -> v

library_context: "L:" library_items
library_items: library_item ("," library_item)*
library_item: HURUF_DIGIT

framework_context: "F:" framework_items
framework_items: framework_item ("," framework_item)*
framework_item: HURUF_DIGIT

tools_context: "T:" tools_items
tools_items: tools_item ("," tools_item)*
tools_item: HURUF_DIGIT

item_line: item_starter? item item_separator?
item_separator: ";"
  | item_separator_newline+
  |
field_method_separator: "|"
statement_separator: ";"
param_separator: ","

item_starter: item_tab+
item_tab: "\\\\t"
item_separator_newline: "\\\\n"

item: package_item
  | import_item
  | export_item
  | block_item
  | statement_item
  | expression_item
  | declarative_item

{package_item}

{import_item}

{export_item}

block_item: scope_item
  | class_item
  | function_item
  | exception_item
  | interface_item

{scope_item}
{class_item}
{function_item}
{exception_item}
{interface_item}

{dataops_statement}

// statement_item dahului block_item
// agar for()body jadi for_statement bukan fungsi for
{statement_item}

{keywords}

declarative_item: "D"

named_values: named_value ("," named_value)*
named_value: nama_identifier tipe_identifier? "=" nilai_identifier
nilai_identifier: expression_item

// ini bisa expression: pemanggilan fungsi (berarti bisa ada dot dong)
nama_tanpa_dot: HURUF_DIGIT_ONLY
nama_identifier: HURUF_DIGIT
nama_identifier_or_literal: literal
  | nama_identifier

// utk bedakan dg nama_identifier, tipe dibikin wajib di sini
// lihat juga: nama_identifier_with_typeparams
nama_jenis_identifier: nama_identifier tipe_identifier
nama_jenis_identifier_optional: nama_identifier tipe_identifier?

// kita hrs punya bentuk: (satu:i, dua?:f)
nama_jenis_identifier_nullable: nama_identifier "?" tipe_identifier

// string, int, float, bool, char, any/object, void
// tipe_data_collection mendahului tipe_data_buatan spy gak kemakan
tipe_identifier: ":" tipe_data_config? tipe_data_builtin 
  | ":" tipe_data_config? tipe_data_collection
  | ":" tipe_data_config? tipe_data_buatan
  | ":" tipe_data_fungsi
  | ":" type_intersection
  | ":" type_union

// sementara hanya built in?
type_intersection: tipe_data_builtin ("&" tipe_data_builtin)+
type_union: tipe_data_builtin ("|" tipe_data_builtin)+

tipe_data_config: "[" tipe_data_config_items "]"
tipe_data_config_items: tipe_data_config_item ("," tipe_data_config_item)*
// dibungkus dg Promise, Option, Result, dll
// gunakan versi boxing: Integer, String, dll
// extend another type?
// juga utk gunakan subtype, i -> long, f -> double, s -> max length if applicable
tipe_data_config_item: "tdci"

tipe_data_builtin: "s" -> string
  | "i" -> integer    // i32
  | "f" -> float
  | "a" -> any        // kita bikin default: undefined
  | "b" -> boolean
  | "c" -> char
  | "dd" -> double
  | "dt" -> datetime
  | "h" -> short      // i16
  | "l" -> long       // i64
  | "y" -> byte       // ~ i8
  | "v" -> void       // any dan object masuk sini, kita bikin default: null
tipe_data_buatan: nama_identifier

tipe_data_semua: tipe_data_builtin
  | tipe_data_collection
  | tipe_data_fungsi
  | tipe_data_buatan
  | object_type

// name: ^&*[nama_identifier:type/type]
// [nama:s/s], ^[nama:s/s], &[nama:s/s], *[nama:s/s]
//tipe_data_fungsi: tipe_data_fungsi_config? "[" nama_identifier ":" tipe_data_semua "/" tipe_data_semua "]"
// apa perlu condition_body? utk statement list?
// const last = (arr:Array<number) => ...condition body... = return arr[arr.length-1]
// %last=[]...condition body...
tipe_data_fungsi: tipe_data_fungsi_config? type_parameters? "[" paramlist* "/" tipe_data_kembali? "]" condition_body?
tipe_data_kembali: tipe_data_semua
tipe_data_fungsi_config: "^" -> function_type_expression // default
  | "&" -> call_signature
  | "*" -> construct_signature

// (nama_identifier: real_type) => real_type
// function_type_expression: 

// hrs bs specify array of what...
// default = array of any/object
// A, D, DF, I, F, N, O, P, S, L
tipe_data_collection: "A" item_type?  -> array      // As
  | "D" (key_type "," value_type)?    -> dict       // Di,s
  | "DF" row_type "," column_type     -> dataframe  // DFs,s
  | "I"                               -> directory  // I
  | "F" file_type?                    -> file       // Fj
  | "N" network_type?                 -> network    // Ngrpc
  | "O" orm_type?                     -> orm        // Osqal
  | "P" (key_type "," value_type)?    -> pair       // Pi,s
  | "S" item_type?                    -> set        // Ss
  | "L" (item_type ("," item_type)*)? -> tuple      // Li,s,f
  | "O" some_type "," none_type          -> optional
  | "R" ok_type "," err_type             -> result
  | "P" value_type                    -> promise

// T, U, V suka digunakan utk generic, jadi tuple = L

// sementara baru tipe data primitif...
item_type: tipe_data_builtin
  | "/" tipe_data_buatan
key_type: tipe_data_builtin
value_type: tipe_data_builtin
  | "/" tipe_data_buatan
row_type: tipe_data_builtin
column_type: tipe_data_builtin
some_type: tipe_data_builtin
none_type: tipe_data_builtin
ok_type: tipe_data_builtin
err_type: tipe_data_builtin

file_type: "t"    -> file_text
  | "c"           -> file_csv
  | "doc"         -> file_doc
  | "h"           -> file_html
  | "j"           -> file_json
  | "rss"         -> file_rss
  | "xls"         -> file_xls
  | "xml"         -> file_xml
  | "y"           -> file_yaml

network_type: "R" -> network_rest
  | "d"           -> network_database
  | "gql"         -> network_graphql
  | "grpc"        -> network_grpc
  | "h"           -> network_http
  | "k"           -> network_kafka  
  | "m"           -> network_mqtt
  | "rmq"         -> network_rabbitmq
  | "s"           -> network_spark
  | "t"           -> network_tcp
  | "u"           -> network_udp
  | "w"           -> network_websocket

orm_type: "dj"    -> orm_django
  | "p"           -> orm_prisma
  | "sqal"        -> orm_sqlalchemy
  | "torm"        -> orm_typeorm

type_parameters: "<" type_parameter_list ">"
type_parameter_list: type_parameter ("," type_parameter)*
type_parameter: nama_identifier constraint?
  | type_parameters

// tipe data collection sementara diskip di sini
constraint: ":" tipe_data_buatan // <satu:dua> => <satu extends dua>
  | ":" tipe_data_builtin
  | ":" tipe_data_fungsi
  | ":" object_type

object_type: "{{" object_contents* "}}"
object_contents: object_content (statement_separator object_content)*
object_content: nama_identifier tipe_identifier

// aslinya "extends" union/intersection/primary-type | functype | ctortype | typegeneric | stringliteral
// class myclass<T:[name:s/s]>
// class myclass<T:{{length:number}}>
nama_identifier_with_typeparams: nama_identifier type_parameters?

{searchable}

{huruf}

number_literal: DEC_NUMBER | HEX_NUMBER | BIN_NUMBER | OCT_NUMBER | FLOAT_NUMBER | IMAG_NUMBER
//string_literal: STRING | LONG_STRING
// Import terminals from standard library (grammars/python.lark)
//%import python (NAME, COMMENT, STRING, LONG_STRING)
%import python (DEC_NUMBER, HEX_NUMBER, OCT_NUMBER, BIN_NUMBER, FLOAT_NUMBER, IMAG_NUMBER)

{base_grammar}
"""



from app.transpiler.zgenerate.refactor.handlers import expression_item
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def string_operation(tree, identifier, language="py"):
    kembali = ""

    for item in anak(tree):
        if data(item) == "penanda":
            pass
        elif data(item) == "concat":
            """
            "+" expression_item
            """
            ei1 = child1(item)
            hasil1 = expression_item.expression_item(ei1, language=language)
            kembali += f"{identifier} + {hasil1}"
        elif data(item) == "char_at":
            """
            /5
            str?s/5
            """
            nilai = token(item)
            # kembali += f'{identifier}[{nilai}]'
            kembali += f"{identifier}.charAt({nilai})"
        elif data(item) == "zfill":
            """
            0...5
            """
            nilai = token(item)
            kembali += f"{identifier}.padStart({nilai}, 0)"
        elif data(item) == "fillz":
            """
            0...5
            """
            nilai = token(item)
            kembali += f"{identifier}.padEnd({nilai}, 0)"
        elif data(item) == "ends_with":
            """
            >hurufdigit
            """
            nilai = token(item)
            kembali += f"{identifier}.endsWith({nilai})"
        elif data(item) == "starts_with":
            """
            <hurufdigit
            """
            nilai = token(item)
            kembali += f"{identifier}.startsWith({nilai})"
        elif data(item) == "contains":
            """
            ~hurufdigit
            """
            nilai = token(item)
            # juga (nilai, startingpos)
            kembali += f"{identifier}.includes({nilai})"
        elif data(item) == "count":
            """
            arr?A#item
            arr?A#42
            """
            nilai = token(item)
            kembali += f'{identifier}.count("{nilai}")'
        elif data(item) == "length":
            """
            ?s|
            """
            kembali += f"{identifier}.length"
        elif data(item) == "search_index":
            """
            s/hurufdigit
            s/kuda
            """
            nilai = token(item)
            # kembali += f'{identifier}.index({nilai})'
            kembali += f"{identifier}.find({nilai})"
        elif data(item) == "last_search_index":
            """
            ls/hurufdigit
            ls/kuda
            """
            nilai = token(item)
            # kembali += f'{identifier}.index({nilai})'
            kembali += f"{identifier}.rfind({nilai})"
        elif data(item) == "index_of":
            """
            i/hurufdigit
            i/kuda
            """
            nilai = token(item)
            kembali += f"{identifier}.indexOf({nilai})"
            # kembali += f'{identifier}.find({nilai})'
        elif data(item) == "last_index_of":
            """
            li/hurufdigit
            https://stackoverflow.com/questions/9572490/find-index-of-last-occurrence-of-a-substring-in-a-string
            when the substring is not found, rfind() returns -1 while rindex() raises an exception ValueError (Python2 link: ValueError).
            jadi rfind lebih aman...
            """
            nilai = token(item)
            kembali += f"{identifier}.lastIndexOf({nilai})"
            # kembali += f'{identifier}.rfind({nilai})'
        elif data(item) == "is_digit":
            """
            9?
            beda isnumeric dan isdigit
            https://www.w3schools.com/python/ref_string_isnumeric.asp
            https://www.w3schools.com/python/ref_string_isdigit.asp
            """
            kembali += f"{identifier}.isdigit()"
        elif data(item) == "is_empty":
            """
            0?
            """
            kembali += f"not {identifier}"
        elif data(item) == "is_space":
            """
            spc?
            """
            kembali += f"{identifier}.isspace()"
        elif data(item) == "to_chars":
            """
            2i
            Array.from('foo'))
            """
            kembali += f'{identifier}.split("")'
        elif data(item) == "to_int":
            """
            2i
            """
            kembali += f"int({identifier})"
        elif data(item) == "to_float":
            """
            2f
            """
            kembali += f"float({identifier})"
        elif data(item) == "to_byte":
            """
            2b
            """
            kembali += f'{identifier}.encode("utf-8")'
        elif data(item) == "from_byte":
            """
            4b
            """
            kembali += f'{identifier}.decode("utf-8")'
        elif data(item) == "to_array":
            """
            "2A"
            ?s2A utk iden.split()
            """
            kembali += f"{identifier}.split()"
        elif data(item) == "to_lines":
            """
            "2A"
            ?s2A utk iden.split()
            """
            kembali += f"{identifier}.splitlines()"
        elif data(item) == "split":
            """
            "sp" "/" HURUF_SYSTEM "/" BILBUL?
            """
            delim = token(item)
            if jumlahanak(item) == 2:
                jumlahsplit = token(item, 1)
                kembali += f'{identifier}.split("{delim}", {jumlahsplit})'
            else:
                kembali += f'{identifier}.split("{delim}")'
        elif data(item) == "trim":
            """
            00
            """
            kembali += f"{identifier}.trim()"
        elif data(item) == "ltrim":
            """
            0.
            """
            kembali += f"{identifier}.ltrim()"
        elif data(item) == "rtrim":
            """
            .0
            """
            kembali += f"{identifier}.rtrim()"
        elif data(item) == "remove_prefix":
            """
            0. huruf
            """
            nilai = token(item)
            kembali += f"{identifier}.removeprefix({nilai})"
        elif data(item) == "remove_suffix":
            """
            .0 huruf
            """
            nilai = token(item)
            kembali += f"{identifier}.removesuffix({nilai})"

        elif data(item) == "upper":
            """
            u
            """
            kembali += f"{identifier}.toUpperCase()"
        elif data(item) == "lower":
            """
            l
            """
            kembali += f"{identifier}.toLowerCase()"
        elif data(item) == "capitalize":
            """
            c
            """
            kembali += f"{identifier}.capitalize()"
        elif data(item) == "slice_with_endpos":
            awal = 0
            akhir = ""  # sampai end
            awal = token(item)
            if jumlahanak(item) == 2:
                # sub5/10 kita pengen inklusif 10
                akhir = token(item, 1)  # tdk spt child, token pake index dari 0
                # akhir = chdata(item, n=2)
                if akhir.isdigit():
                    akhir = int(akhir) + 1
            if akhir:
                kembali += f"{identifier}.slice({awal}, {akhir})"
            else:
                kembali += f"{identifier}.slice({awal})"
        elif data(item) == "substring_with_endpos":
            """
            "ub" BILBUL (":" BILBUL)? -> substring_with_endpos // .substring
            substring(start, end)
            string_operation
              penanda
              substring_with_endpos     5
            mystr?sub5
            """
            # from pprint import pprint
            # pprint(item)
            awal = 0
            akhir = ""  # sampai end
            awal = token(item)
            if jumlahanak(item) == 2:
                # sub5/10 kita pengen inklusif 10
                akhir = token(item, 1)  # tdk spt child, token pake index dari 0
                # akhir = chdata(item, n=2)
                if akhir.isdigit():
                    akhir = int(akhir) + 1
            if akhir:
                kembali += f"{identifier}.substring({awal}, {akhir})"
            else:
                kembali += f"{identifier}.substring({awal})"
        elif data(item) == "substring_with_length":
            """
            "ub" BILBUL "/" BILBUL    -> substring_with_length // .substr
            substr(start, len)
            ?sub dg / berarti with panjang -> .substr
            ?sub dg : berarti with endpos -> .substring
            """
            awal = token(item)
            panjang = token(item, 1)
            result = int(awal) + int(panjang)
            kembali += (
                f"{identifier}.substr({awal}, {result})"  # kurang cantik tapi jelas
            )
        elif data(item) == "replace":
            """
            "r" "/" HURUF_DIGIT "/" HURUF_DIGIT "/"
            """
            old = child1(item)
            old = expression_item.expression_item(old, language=language)
            new = child2(item)
            new = expression_item.expression_item(new, language=language)
            kembali += f"{identifier}.replace({old}, {new})"
        elif data(item) == "center":
            """
            "_" BILBUL ("/" HURUF_SYSTEM)? -> center
            https://www.w3schools.com/python/ref_string_center.asp
            """
            lebar = 0
            karakter = ""
            lebar = token(item)
            if jumlahanak(item) == 2:
                karakter = token(item, 1)
            if karakter:
                kembali += f'{identifier}.center({lebar}, "{karakter}")'
            else:
                kembali += f"{identifier}.center({lebar})"
        elif data(item) == "ljust":
            """
            "_" BILBUL ("/" HURUF_SYSTEM)? -> center
            https://www.w3schools.com/python/ref_string_center.asp
            https://www.w3schools.com/python/ref_string_ljust.asp
            """
            lebar = 0
            karakter = ""
            lebar = token(item)
            if jumlahanak(item) == 2:
                karakter = token(item, 1)
            if karakter:
                kembali += f'{identifier}.ljust({lebar}, "{karakter}")'
            else:
                kembali += f"{identifier}.ljust({lebar})"
        elif data(item) == "format":
            """
            ?sfm:name=usef,gf=wieke
            format
              named_values
                named_value
                  nama_identifier     name
                  nilai_identifier
                    expression_item
                      nama_identifier usef
                named_value
                  nama_identifier     gf
                  nilai_identifier
                    expression_item
                      nama_identifier wieke
            https://www.w3schools.com/python/ref_string_format.asp
            txt.format(price = 49)
            """
            named_values = child1(item)
            kembali += f"{identifier}.format({named_values})"
        elif data(item) == "translate":
            """
            txt?str:abc=ABC

            https://www.w3schools.com/python/ref_string_maketrans.asp
            https://www.w3schools.com/python/ref_string_translate.asp
            mystr.translate(mystr.maketrans(pertama, kedua))
            string_operation
              penanda
              translate <- mulai dari sini
                abc <- token pertama
                ABC <- token kedua
            """
            pertama = token(item)
            kedua = token(item, 1)
            kembali += (
                f"{identifier}.translate({identifier}.maketrans({pertama}, {kedua}))"
            )

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



# X:(App)
# X:conf1,conf2/(App)
export_item = """
export_item: keyword_export export_config? export_content

// export src/calc.dart, src/music.dart
// export "src/calc.dart"; newline export "src/music.dart"

export_config: exportconfiglist "/"
exportconfiglist: export_config_item+
export_config_item: "d" -> export_default
  | "m" -> export_module    // module.exports = .., correspondingly utk import const .. = require(..)
  | "o" -> export_objectify // export {bla1, bla2, bla3};
export_content: "(" exportcontentlist ")"
exportcontentlist: exportcontent ("," exportcontent)*
exportcontent: HURUF_EXPORT
"""



from app.fmus import Fmus
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)
from app.utils import env_int, tidur


def fmus_run(program, language="py"):
    kembali = ""
    fmus = Fmus(env_int("ULIBPY_FMUS_DEBUG"))
    # fmus.set_file_dir_template(filepath)
    # fmus.set_file_template(filepath)
    # fmus.set_dir_template_from_file(filepath)
    if env_int("ULIBPY_FMUS_CAPTURE_STDOUT_STDERR"):
        fmus.process(program, capture_outerr=True)
        if "$*" in program:
            tidur(ms=env_int("ULIBPY_STDOUT_CAPTURE_SLEEP_MS"))
            if fmus.stdout or fmus.stderr:
                kembali += fmus.stdout if fmus.stdout else fmus.stderr
    else:
        fmus.process(program)

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import (
    function_content, function_param, nama_identifier_with_typeparams,
    tipe_identifier)
from app.transpiler.zgenerate.refactor.handlers.common import (dec, inc,
                                                               self_tab)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def method_content(tree, language="py"):
    kembali = ""
    _tipe_identifier = ""
    for item in anak(tree):
        if data(item) == "function_name":
            _nama_identifier_with_typeparams = child1(item)
            nama = nama_identifier_with_typeparams.nama_identifier_with_typeparams(
                _nama_identifier_with_typeparams, language=language
            )
            kembali += f"{nama}"
        elif data(item) == "tipe_identifier":
            _tipe_identifier = tipe_identifier.tipe_identifier(item, language=language)
        elif data(item) == "function_param":
            kembali += function_param.function_param(item, language=language)
            # tipe fungsi stlh param
            if _tipe_identifier:
                kembali += f": {_tipe_identifier}"
        elif data(item) == "function_content":
            kembali += " {\n"
            inc()
            kembali += function_content.function_content(item, language=language)
            kembali += "\n"
            dec()
            kembali += self_tab() + "}"

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import (declaration_value,
                                                        tipe_identifier)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def field_content(tree, language="py"):
    kembali = ""

    fieldconfig, fieldname, fieldtype, fieldvalue = "", "", "", ""
    for item in anak(tree):
        if data(item) == "field_config":
            pass
        elif data(item) == "field_name":
            fieldname = token(item)
            # print('\t nama =', fieldname)
        elif data(item) == "tipe_identifier":
            fieldtype = tipe_identifier.tipe_identifier(item, language=language)
            # print('\t jenis =', fieldtype)
        elif data(item) == "declaration_value":
            fieldvalue = declaration_value.declaration_value(item, language=language)
            # print('\t nilai =', fieldvalue)
    if fieldconfig:
        kembali += fieldconfig + " "
    kembali += fieldname
    if fieldtype:
        kembali += f": {fieldtype}"
    if fieldvalue:
        kembali += " = " + fieldvalue

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import (
    anonymous_function, expression_item, named_values,
    stringify_anonymous_function_result)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def argument_list(tree, language="py"):
    kembali = ""

    args = []
    for arg in anak(tree):
        # argument
        for tipearg in anak(arg):
            if data(tipearg) == "named_values":
                # print('named_values #1')
                hasil = named_values.named_values(tipearg, language=language)
                # print('named_values #2', hasil)
                args.append(hasil)
            elif data(tipearg) == "expression_item":
                hasil = expression_item.expression_item(tipearg, language=language)
                hasil = str(hasil)
                args.append(hasil)
            elif data(tipearg) == "anonymous_function":
                hasil = anonymous_function.anonymous_function(
                    tipearg, language=language
                )
                hasil = stringify_anonymous_function_result.stringify_anonymous_function_result(
                    hasil, language=language
                )
                # kembalian AF adlh dict
                # if isinstance(hasil, dict):
                #   if 'external' in hasil and not hasil['external'] and 'internal' in hasil:
                #     hasil = hasil['internal']
                #   elif 'internal' in hasil and not hasil['internal']:
                #     hasil = hasil['external']
                args.append(hasil)

    kembali = args  # list

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import (function_config,
                                                        function_content,
                                                        function_name,
                                                        function_param,
                                                        tipe_identifier)
from app.transpiler.zgenerate.refactor.handlers.common import (dec, inc,
                                                               tabify_content)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def function_item(tree, language="py"):
    print("function_item")
    kembali = ""
    fnkw = ""
    fnconfig, fnname, fntype, fnparam, fncontent = "", "", "", "", ""
    is_anonymous = False

    for item in anak(tree):
        print("function_item anak:", data(item))
        if data(item) == "keyword_function":
            fnkw = "function"
        elif data(item) == "function_config":
            fnconfig = function_config.function_config(item, language=language)
        elif data(item) == "function_name":
            print(
                "mau panggil function name, lihat tipe:",
                type(function_name),
                type(item),
                type(language),
            )
            fnname = function_name.function_name(item, language=language)
            # kembali += 'function ' + name
        elif data(item) == "tipe_identifier":
            fntype = tipe_identifier.tipe_identifier(item, language=language)
        elif data(item) == "function_param":
            fnparam = function_param.function_param(item, language=language)
        elif data(item) == "function_content":
            inc()
            fncontent = function_content.function_content(item, language=language)
            dec()

    if language == "py":
        pass
    elif language == "ts":
        if fnconfig and "anonymous" in fnconfig:
            is_anonymous = True
            fnconfig.remove("anonymous")
        if fnconfig and "arrow" in fnconfig:
            fnconfig.remove("arrow")
            if "iife" in fnconfig:
                fnconfig.remove("iife")
                if fnconfig:
                    fnconfig = " ".join(fnconfig) + " "
                # jk iife maka gak perlu fnname dan fntype
                # (async () => {})()
                kembali += f"({fnconfig}{fnparam} => {{\n{fncontent}\n}})()"
            else:
                if fnconfig:
                    # print('#4', fnconfig, type(fnconfig))
                    kembali += " ".join(fnconfig) + " "
                if is_anonymous:
                    # (req, res) => {}
                    kembali += f"{fnparam}"
                else:
                    kembali += f"const {fnname} = {fnparam}"
                if fntype:
                    kembali += fntype
                kembali += " => {\n" + fncontent + "\n}"
        else:
            if "iife" in fnconfig:
                fnconfig.remove("iife")
                # print('#5', fnconfig, type(fnconfig))
                if fnconfig:
                    fnconfig = " ".join(fnconfig) + " "
                # jk iife maka gak perlu fnname dan fntype
                # (async () => {})()
                kembali += f"({fnconfig}function{fnparam} {{\n{fncontent}\n}})()"
            else:
                if fnconfig:
                    fnconfig = " ".join(fnconfig)
                    kembali += fnconfig + " "
                if is_anonymous:
                    # function (req, res) {}
                    kembali += f"{fnkw} {fnparam}"
                else:
                    kembali += f"{fnkw} {fnname}{fnparam}"
                if fntype:
                    kembali += fntype
                kembali += " {\n" + fncontent + "\n}"
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



if_statement = """
// if (if) {var myvar = 24}
if_statement: if_keyword condition_then_if (condition_then_elif)* condition_then_else?

condition_then_if: condition_if condition_body
condition_then_elif: condition_elif condition_body
condition_then_else: condition_else condition_body

condition_if: "(" expression_item ")"
condition_elif: "'(" expression_item ")"
condition_else: "()"
"""



from app.transpiler.zgenerate.refactor.handlers import (
    anonymous_function, arithmetic_expression, casting_expression,
    expression_item, function_call, literal, member_dot_expression,
    member_index_expression, range_expression, relational_expression)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def ternary_expression(tree, language="py"):
    """
    /ts/i>42?'lebih dari 42':'kurang dari samadengan 42'
    /py/i>42?'lebih dari 42':'kurang dari samadengan 42'
    """
    print("ternary_expression")
    kembali = ""
    ei_condition = child1(tree)
    ei_condition = expression_item.expression_item(ei_condition, language=language)
    ei_yes = child2(tree)
    ei_yes = expression_item.expression_item(ei_yes, language=language)
    ei_no = child3(tree)
    ei_no = expression_item.expression_item(ei_no, language=language)
    kembali = f"{ei_condition} ? {ei_yes} : {ei_no}"

    if language == "py":
        kembali = f"{ei_yes} if {ei_condition} else {ei_no}"
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass

    return kembali



# :xwi/myfunc:i(){}
function_item_header = """
function_item: keyword_function function_config? function_name tipe_identifier? function_param function_content

function_config: functionconfiglist "/"

//functionconfiglist: functionconfig ("," functionconfig)*
//:ai/foo(){}
functionconfiglist: functionconfig+

// language config -> /js, /py, /kt, /rs, /scala
// decorator
// async
// iife
// anonymous, lambda, closure, inner, arrow
// contoh di rust: const myconst = async (funcparams): Promise<sometype> => {}
// export + default
functionconfig: "+" -> public
  | "-" -> private
  | "#" -> protected
  | "%" -> static
  | "a" -> async
  | "x" -> export
  | "w" -> arrow
  | "i" -> iife
  | "no" -> anonymous

function_name: nama_identifier_with_typeparams

function_type: ":" HURUF_DIGIT

function_param: "(" paramlist* ")"
paramlist: param ("," param)*
// nama_jenis_identifier_optional = nama_identifier | nama_jenis_identifier
param: named_values
  | nama_identifier
  | nama_jenis_identifier
  | nama_jenis_identifier_nullable

// digunakan di expression-item:
// new_operator nama_identifier function_call_param -> instantiation_expression
function_call_param: "(" callparamlist* ")"
callparamlist: callparam ("," callparam)*
callparam: named_values
  | nama_identifier
  | literal  

// ini jadi setara dg function_call_param
argument_list: argument ("," argument)*
argument: expression_item
  | named_values
  | anonymous_function

function_content: "{" functioncontentlist* "}"
functioncontentlist: functioncontent (statement_separator functioncontent)*
functioncontent: statement_item
"""

from .anonymous_function import anonymous_function

function_item = f"""
{function_item_header}

{anonymous_function}
"""



# >(){}
constructor_content = """

// > utk bedakan dg method
constructor_content: ">" constructor_config? function_param function_content

constructor_config: "[" constructorconfiglist "]"
constructorconfiglist: constructorconfig ("," constructorconfig)*
// termasuk ada destructor
constructorconfig: "c"
"""



from app.transpiler.zgenerate.refactor.handlers import (for_each, for_in,
                                                        for_of,
                                                        for_traditional)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def for_variation(tree, language="py"):
    kembali = ""

    varian = child1(tree)
    if data(varian) == "for_traditional":
        # hasil = for_traditional(varian)
        # output += hasil
        kembali = for_traditional.for_traditional(varian, language=language)
    elif data(varian) == "for_each":
        kembali = for_each.for_each(varian, language=language)
    elif data(varian) == "for_in":
        kembali = for_in.for_in(varian, language=language)
    elif data(varian) == "for_of":
        kembali = for_of.for_of(varian, language=language)
    else:
        kembali = ""
    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



string_operation = """

string_operation:       penanda "s" string_operation_choice
// kita mau bikin adding newline/tab/dll segampang mungkin
string_operation_choice: "+" expression_item -> concat       // mystr?s+ ...
  | "/" BILBUL          -> char_at      // mystr?s/5
  | ">" HURUF_DIGIT     -> ends_with
  | "<" HURUF_DIGIT     -> starts_with
  | "~" HURUF_DIGIT     -> contains
  | "#" HURUF_DIGIT     -> count
  | "|"                 -> length
  | "s/" HURUF_DIGIT    -> search_index    // mystr?ss/cari
  | "ls/" HURUF_DIGIT   -> last_search_index
  | "i/" HURUF_DIGIT    -> index_of     // mystr?si/5
  | "li/" HURUF_DIGIT   -> last_index_of
  | "9?"                -> is_digit
  | "0?"                -> is_empty
  | "spc?"              -> is_space
  | "2c"                -> to_chars
  | "2i"                -> to_int
  | "2f"                -> to_float
  | "2b"                -> to_byte
  | "4b"                -> from_byte
  | "2l"                -> to_literal
  | "4l"                -> from_literal
  | "2A" -> to_array     // 'hello'.split() = default by space
  | "2L" -> to_lines
  | "sp" "/" HURUF_NON_SLASH "/" BILBUL?  -> split
  | "r" "/" expression_item "/" expression_item "/"? -> replace // mystr?sr/cari/ganti/
  | "0..." BILBUL         -> zfill
  | "...0" BILBUL         -> fillz
  | "00"                 -> trim         // mystr?s0
  | "0."                 -> ltrim
  | ".0"                 -> rtrim
  | "0." HURUF_SYSTEM                -> remove_prefix
  | ".0" HURUF_SYSTEM                -> remove_suffix
  | "u"                 -> upper        // mystr?su
  | "l"                 -> lower        // mystr?sl
  | "c"                 -> capitalize   // mystr?sc
  | "ub" BILBUL (":" BILBUL)? -> substring_with_endpos // .substring
  | "ub" BILBUL "/" BILBUL    -> substring_with_length // .substr
  | "li" BILBUL (":" BILBUL)? -> slice_with_endpos
  | "__" BILBUL ("/" HURUF_SYSTEM)? -> center
  | "<<" BILBUL ("/" HURUF_SYSTEM)? -> ljust
  | "fm:" named_values -> format // ?sfm:name=usef,gf=wieke
  | "tr:" HURUF_DIGIT "=" HURUF_DIGIT -> translate // 
  | "ma:" HURUF_DIGIT ("/" regex_match_configs)? -> regex_match

HURUF_NON_SLASH: 					(" "|LETTER|DIGIT|"*"|"."|"\\\\") 	(LETTER|DIGIT|"_"|"*"|"."|"-"|"+"|" "|":"|"\\"|"\"")*

regex_match_configs: regex_match_configs+
regex_match_config: "g" -> regex_match_global
  | "i" -> regex_match_case_insensitive

// perlu juga method: char_in_string -> if c in s
// https://youtu.be/LK7tDnzUZZM?t=271 s.find(c) != string::npos
"""



from app.transpiler.zgenerate.refactor.handlers import expression_item
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def dict_operation(tree, identifier, language="py"):
    kembali = ""

    for item in anak(tree):
        if data(item) == "penanda":
            pass
        elif data(item) == "concat_extend_update":
            """
            "+=" expression_item
            concat_extend_update
              expression_item
                literal
                  literal_dict
                    dict_items
                      dict_item
                        dict_item_name        a
                        expression_item
                          literal
                            literal_number    1
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.update({hasil})"
        elif data(item) == "get_item":
            """
            "/" literal ("/" literal)?
            iden.get(item, item)
            get_item
              expression_item
                nama_identifier       item
              expression_item
                nama_identifier       default
            """
            ei = child1(item)
            hasil1 = expression_item.expression_item(ei, language=language)
            hasil2 = hasil1
            if jumlahanak(item) == 2:
                ei2 = child2(item)
                hasil2 = expression_item.expression_item(ei2, language=language)
            kembali += f"{identifier}.get({hasil1}, {hasil2})"
        elif data(item) == "length":
            """
            dik?D|
            """
            kembali += f"Object.keys({identifier}).length"
        elif data(item) == "assign":
            """
            "+" expression_item "=" expression_item
            dik[item]=nilai
            dik?D+item=nilai
            """
            ei1 = child1(item)
            ei2 = child2(item)
            hasil1 = expression_item.expression_item(ei1, language=language)
            hasil2 = expression_item.expression_item(ei2, language=language)
            kembali += f"{identifier}[{hasil1}] = {hasil2}"
        elif data(item) == "has_item":
            """
            "~(" expression_item ")"
            """
            ei1 = child1(item)
            hasil1 = expression_item.expression_item(ei1, language=language)
            kembali += f"{hasil1} in {identifier}.values()"
        elif data(item) == "remove_item":
            """
            "-(" expression_item ")"
            filtereddik = {k:v for k,v in identifier.items() if v != hasil}
            """
            ei1 = child1(item)
            hasil1 = expression_item.expression_item(ei1, language=language)
            kembali += f"{{k:v for k,v in {identifier}.items() if v != {hasil1}}}"
        elif data(item) == "has_key":
            """
            "~" expression_item
            """
            ei1 = child1(item)
            hasil1 = expression_item.expression_item(ei1, language=language)
            # kembali += f'{hasil1} in {identifier}'
            kembali += f"{identifier}.includes({hasil1})"
        elif data(item) == "remove_key":
            """
            "-" expression_item
            """
            ei1 = child1(item)
            hasil1 = expression_item.expression_item(ei1, language=language)
            kembali += f"del {identifier}[{hasil1}]"
        elif data(item) == "clear":
            """
            dik?D0
            """
            kembali += f"{identifier}.clear()"
        elif data(item) == "remove_at_tail":
            """
            popitem = remove last inserted
            dik?D->
            """
            kembali += f"{identifier}.popitem()"
        elif data(item) == "remove_at_index":
            """
            dik?D-@5
            """
            hasil = token(item)
            kembali += f"{identifier}.pop({hasil})"
        elif data(item) == "is_empty":
            """
            dik?D0?
            """
            # kembali += f'len({identifier}) == 0'
            kembali += identifier
        elif data(item) == "entries":
            """
            dik?D@
            """
            kembali += f"{identifier}.entries()"
        elif data(item) == "keys":
            """
            dik?D#
            """
            kembali += f"{identifier}.keys()"
        elif data(item) == "values":
            """
            dik?D$
            """
            kembali += f"{identifier}.values()"
    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.dirutils import here, joiner
from app.transpiler.zgenerate.refactor.handlers import (import_things,
                                                        searchable)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def import_item(tree, language="py"):
    kembali = ""
    container = None
    merged_imports = []
    search_mode = False
    hasil = ""
    filename = f"import_{language}.txt"
    for item in anak(tree):
        if data(item) == "import_container":
            container = token(item)
        elif data(item) == "import_things":
            hasil += import_things.import_things(item, language=language)
            merged_imports.append(hasil)
        elif data(item) == "searchable":
            search_mode = True
            filepath = joiner(here(__file__), "../../../data/", filename)
            print("filepath:", filepath)
            kembali += searchable.searchable(item, filepath, language=language)

    if search_mode:
        pass  # hasil sudah berisi yg diminta
    else:
        if container:
            kembali += f'import {", ".join(merged_imports)} from {container};'
        else:
            kembali += f'import {", ".join(merged_imports)};'

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import object_content
from app.transpiler.zgenerate.refactor.handlers.common import (dec, inc,
                                                               tabify_content)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def object_type(tree, language="py"):
    kembali = ""

    if not beranak(tree):
        return "{}"
    object_contents = child1(tree)
    contents = []
    for item in anak(object_contents):
        # ingat ada juga content = statement_separator
        if data(item) == "object_content":
            buah = object_content.object_content(item, language=language)
            contents.append(buah)

    kembali = "\n".join(contents)
    inc()
    kembali = tabify_content(kembali)
    dec()
    kembali = "{\n" + kembali + "\n}"

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import (
    anonymous_function, array_map, expression_item, initialize_with_faker,
    stringify_anonymous_function_result)
from app.transpiler.zgenerate.refactor.handlers.common import (dec, inc,
                                                               self_tab)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def array_operation(tree, identifier, language="py"):
    kembali = ""

    for item in anak(tree):
        """
        array_filter, array_reduce
        """
        if data(item) == "penanda":
            pass
        elif data(item) == "item_at":
            # lst?A/5
            at = token(item)
            kembali += f"{identifier}[{at}]"
        elif data(item) == "item_at_set":
            # lst?A/5=nilai_baru
            at = token(item)
            ei = child2(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}[{at}] = {hasil}"
        elif data(item) == "index_of":
            nama_identifier_or_literal = child1(item)
            hasil = nama_identifier_or_literal(
                nama_identifier_or_literal, language=language
            )
            # kembali += f'{identifier}.index({hasil})'
            kembali += f"{identifier}.indexOf({hasil})"
        elif data(item) == "slice_array":
            slice_startingpos = token(item)
            if jumlahanak(item) == 2:
                slice_endingpos = token(item, 1)
                kembali += f"{identifier}.slice({slice_startingpos}, {slice_endingpos})"
            else:
                kembali += f"{identifier}.slice({slice_startingpos})"
        elif data(item) == "insert_at_tail":
            """
            ini adlh standard push/append
            myvec.push(42) di rust
            mylist.append(42) di py

            lst?A+  ei
            arr?A+42
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.push({hasil})"
        elif data(item) == "concat_extend_update":
            """
            lst?A+= ei
            lst?A+= [42]
            list += [42]
            """
            eis = []
            # ei = child1(item)
            # hasil = expression_item(ei)
            for ei in anak(item):
                thing = expression_item.expression_item(ei, language=language)
                eis.append(thing)
            hasil = ", ".join(eis)
            kembali += f"{identifier}.concat({hasil})"
        elif data(item) == "insert_at_head":
            """
            lst?A+< ei
            lst?A+< 42
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            # atau identifier = [hasil] + identifier
            kembali += f"{identifier}.unshift({hasil})"
        elif data(item) == "insert_at_index":
            """
            insert_at_index
              expression_item
                nama_identifier       item
              5
            insert_at_index
              expression_item
                literal
                  literal_number      42
              5
            lst?A+ ei @5
            lst?A+ myitem @5
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)

            posisi = token(item, 1)  # child no 2 adlh token
            kembali += f"{identifier}.insert({posisi}, {hasil})"
        elif data(item) == "length":
            """
            len(arr)
            lst?A|
            """
            kembali += f"{identifier}.length"
        elif data(item) == "remove_at_index":
            """
            lst?A-5
            del arr[5]
            arr.pop(5)
            """
            posisi = token(item)
            # kembali += f'{identifier}.pop({hasil})'
            kembali += f"delete {identifier}[{hasil}]"
        elif data(item) == "splice_add_remove":
            # "+S-" BILBUL "/" BILBUL "/" expression_item ("," expression_item)*
            start_removepos = token(item)
            howmany_removeitem = token(item, 1)
            eis = []
            for ei in anak(item)[2:]:
                hasil = expression_item.expression_item(ei, language=language)
                eis.append(hasil)
            added_things = ", ".join(eis)
            kembali += f"{identifier}.splice({start_removepos}, {howmany_removeitem}, {added_things})"
        elif data(item) == "splice_remove":
            # "-S-" BILBUL "/" BILBUL
            start_removepos = token(item)
            howmany_removeitem = token(item, 1)
            kembali += f"{identifier}.splice({start_removepos}, {howmany_removeitem})"
        elif data(item) == "pop_item":
            posisi = 0
            # if beranak(item):
            #   posisi = token(item)
            #   kembali += f'{identifier}.pop({posisi})'
            # else:
            kembali += f"{identifier}.pop()"
        elif data(item) == "shift_item":
            # kembali += f'item = {identifier}[0]; del {identifier}[0]'
            kembali += f"{identifier}.shift()"
        elif data(item) == "remove_at_head":
            """
            arr?A-<
            """
            # kembali += f'{identifier}.pop(0)'
            kembali += f"del {identifier}[0]"
        elif data(item) == "remove_at_tail":
            """
            arr?A->
            """
            # kembali += f'{identifier}.pop(-1)'
            # kembali += f'{identifier}.pop()'
            kembali += f"del {identifier}[-1]"
        elif data(item) == "remove_by_value":
            """
            arr?A-ei
            """
            ei = child1(item)
            thing = expression_item.expression_item(ei, language=language)
            # index = identifier.indexOf(item)
            # if (index != -1) { identifier.splice(index, 1) }
            kembali += f"{identifier}.filter(item => item !== {thing})"
        elif data(item) == "min":
            """
            arr?A<<
            """
            kembali += f"Math.min.apply(null, {identifier})"
        elif data(item) == "max":
            """
            arr?A>>
            """
            kembali += f"Math.max.apply(null, {identifier})"
        elif data(item) == "average":
            """
            arr?A==
            """
            kembali += f"round(sum({identifier})/len({identifier}), 2)"
        elif data(item) == "sum":
            """
            arr?A++
            """
            kembali += f"sum({identifier})"
        elif data(item) == "contains":
            """
            arr?A~item
            arr?A~42
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            # kembali += f'{hasil} in {identifier}'
            kembali += f"{identifier}.includes({hasil})"
        elif data(item) == "count":
            """
            arr?A#item
            arr?A#42
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.count({hasil})"
        elif data(item) == "clear":
            """
            arr?A0
            """
            kembali += f"{identifier}.clear()"
        elif data(item) == "is_empty":
            """
            arr?A0?
            """
            kembali += identifier
        elif data(item) == "to_string":
            """
            arr?A2s
            """
            kembali += f"{identifier}.toString()"
        elif data(item) == "split":
            """
            arr?Asp
            arr?Asp/hurufsystem/
            """
            if beranak(item):
                nilai = token(item)
                kembali += f'{identifier}.join("{nilai}")'
            else:
                kembali += f'{identifier}.join(" ")'
        elif data(item) == "reverse":
            """
            arr?A!
            """
            # kembali += f'{identifier} = {identifier}[::-1]'
            # for i in reversed(iterator)
            kembali += f"{identifier}.reverse()"
        elif data(item) == "sort":
            """
            arr?A$
            ada juga numeric sort:
            const points = [40, 100, 1, 5, 25, 10];
            points.sort(function(a, b){return a - b});
            """
            kembali += f"{identifier}.sort()"
        elif data(item) == "reverse_sort":
            """
            arr?A$!
            """
            kembali += f"{identifier}.sort(reverse=True)"
        elif data(item) == "array_map":
            """
            arr?AM#(){}
            """
            kembali = array_map.array_map(item, language=language)
            kembali["operation"] = "map"
        elif data(item) == "array_filter":
            """
            arr?AF#(){}
            """
            kembali = array_map.array_map(
                item, language=language
            )  # sementara belum array_filter
            kembali["operation"] = "filter"
        elif data(item) == "array_reduce":
            """
            arr?AR#(){}
            """
            kembali = array_map.array_map(
                item, language=language
            )  # sementara belum array_reduce
            kembali["operation"] = "reduce"
        elif data(item) == "item_first":
            # "0/" -> item_first
            kembali += f"{identifier}[0]"
        elif data(item) == "item_last":
            # "9/" -> item_last
            kembali += f"{identifier}[-1]"
        elif data(item) == "array_traversal_for_traditional":
            # "44" nama_identifier? -> array_traversal_for_traditional
            iden = "item"
            if beranak(item) and chdata(item) == "nama_identifier":
                iden = chtoken(item)
            inc()
            kembali += (
                f"for index, {iden} in enumerate({identifier}):\n{self_tab()}pass"
            )
            dec()
        elif data(item) == "array_traversal_for_each":
            # "4@" nama_identifier? -> array_traversal_for_each
            iden = "item"
            if beranak(item) and chdata(item) == "nama_identifier":
                iden = chtoken(item)
            inc()
            kembali += (
                f"for index, {iden} in enumerate({identifier}):\n{self_tab()}pass"
            )
            dec()
        elif data(item) == "array_traversal_for_key":
            # "4#" nama_identifier? -> array_traversal_for_key
            iden = "index"
            if beranak(item) and chdata(item) == "nama_identifier":
                iden = chtoken(item)
            inc()
            kembali += f"for {iden},_ in enumerate({identifier}):\n{self_tab()}pass"
            dec()
        elif data(item) == "array_traversal_for_value":
            # "4$" nama_identifier? -> array_traversal_for_value
            iden = "item"
            if beranak(item) and chdata(item) == "nama_identifier":
                iden = chtoken(item)
            inc()
            kembali += f"for {iden} in {identifier}:\n{self_tab()}pass"
            dec()

        elif data(item) == "is_array":
            # myarr?AA?
            kembali += f"Array.isArray({identifier})"
        elif data(item) == "uniqify":
            # myarr?AU
            kembali += f"const {identifier} = Array.from(new Set({identifier}))"

        elif data(item) == "new_array":
            # "n" array_new_operation? -> new_array
            #     "@" BILBUL -> allocate
            #     "$" expression_item -> initialize
            #     "$" BILBUL tipe_identifier -> initialize_with_faker
            # /py/myarr?An@5
            # /py/myarr?An$[1,2,3,4,5]
            for cucu in anak(item):
                jenis = data(cucu)
                if jenis == "allocate":
                    jumlah = token(cucu)
                    # print('jumlah:', jumlah)
                    jumlah = int(jumlah)
                    kembali += f"{identifier} = {[0]*jumlah}"
                elif jenis == "initialize":
                    ei = child1(cucu)
                    hasil = expression_item.expression_item(ei, language=language)
                    print("initialize=>", hasil, type(hasil))
                    kembali += f"{identifier} = {hasil}"
                elif jenis == "initialize_with_faker":
                    hasil = initialize_with_faker.initialize_with_faker(
                        cucu, language=language
                    )
                    kembali += f"{identifier} = {hasil}"
                elif jenis == "create_with_arrayfrom":
                    # lst?AnF(){?+'hello'}
                    mapper_function_for_arraylike = child1(cucu)
                    _anonymous_function = child1(mapper_function_for_arraylike)
                    mapper_func = anonymous_function.anonymous_function(
                        _anonymous_function, language=language
                    )
                    mapper_func = stringify_anonymous_function_result.stringify_anonymous_function_result(
                        mapper_func, language=language
                    )
                    kembali += f"const {identifier} = Array.from(some_arraylike_object, {mapper_func})"
                elif jenis == "allocate_with_arrayfrom":
                    # lst?An@F5/0
                    # buat array sepanjang 5 berisi 0 setara [0,0,0,0,0]
                    panjang = token(cucu)
                    value = token(cucu, 1)
                    kembali += f"const {identifier} = Array.from({{ {panjang} }}, () => {value})"

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali
